
/*
 * Mandrake KWin client based on Keramik kwin client
 *
 * Copyright (C) 2003-2004 Laurent Montel <lmontel@mandrakesoft.com>
 * Copyright (C) 2002 Fredrik Hglund <fredrik@kde.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the license, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <kdebug.h>
#include <kconfig.h>
#include <klocale.h>
#include <kiconeffect.h>

#include <qpainter.h>
#include <qlayout.h>
#include <qbitmap.h>
#include <qstyle.h>
#include <qtooltip.h>
#include <qlabel.h>

#include <X11/Xlib.h>

#include "mandrakewinclient.h"
#include "mandrakewinclient.moc"



namespace Mandrake
{

	const int buttonMargin     =  9;  // Margin between the window edge and the buttons
	const int buttonSpacing    =  2;  // Spacing between the titlebar buttons
	const int iconSpacing      =  5;  // Spacing between the icon and the text label
	const int bottomCornerSize = 30;  // Size of the bottom diagonal resize corners

	// Default button layout
	const char default_left[]  = "M";
	const char default_right[] = "IAX";

	MandrakeHandler *clientHandler = NULL;
	bool mandrake_initialized = false;



// -------------------------------------------------------------------------------------------



MandrakeHandler::MandrakeHandler()
{
	for ( int i = 0; i < NumTiles; i++ ) {
		activeTiles[i]   = NULL;
		inactiveTiles[i] = NULL;
	}
	for ( int i = 0; i < NumButtonDecos; i++ )
		buttonDecos[i]=NULL;

	settings_cache = NULL;

	imageDb = MandrakeImageDb::instance();

	readConfig();
	createPixmaps();
#if 0
	// Selfmask the bitmaps
	for ( int i = 0; i < NumButtonDecos; i++ )
	  if(buttonDecos[i] )
	    buttonDecos[i]->setMask( *buttonDecos[i]->mask() );
#endif

	// Flip the bitmaps horizontally in right-to-left mode
	if ( QApplication::reverseLayout() ) {
		for ( int i = 0; i < Help; i++ )
                    if( buttonDecos[i] )
			flip( reinterpret_cast<QPixmap**>(buttonDecos)[i] );

		for ( int i = Help + 1; i < NumButtonDecos; i++ )
                    if( buttonDecos[i] )
			flip( reinterpret_cast<QPixmap**>(buttonDecos)[i] );
	}

	readConfig();
	createPixmaps();

	mandrake_initialized = true;
}


MandrakeHandler::~MandrakeHandler()
{
	mandrake_initialized = false;
	destroyPixmaps();

	for ( int i = 0; i < NumButtonDecos; i++ )
		delete buttonDecos[i];

        delete settings_cache;

	MandrakeImageDb::release();
	imageDb = NULL;
	clientHandler = NULL;
}


void MandrakeHandler::createPixmaps()
{
  QColor captionColorActive, buttonColor;
  QColor captionColorInactive;
  QImage *titleCenter = NULL,
    *captionRight = NULL, *captionCenter = NULL;

  // Active tiles
  // -------------------------------------------------------------------------
  //don't apply color effect when we use mdk default color
  captionColorActive = KDecoration::options()->color( ColorTitleBar,   true );
  if( captionColorActive == QColor("#21449C"))
    captionColorActive = QColor();

  captionColorInactive = KDecoration::options()->color( ColorTitleBar,   false );
  if(captionColorInactive == QColor("#EDEDED"))
    captionColorInactive = QColor();

  // Load the titlebar corners.
  activeTiles[ TitleLeft ]  = loadPixmap( "titlebar-left",  captionColorActive  );
  activeTiles[ TitleRight ] = loadPixmap( "titlebar-right", QColor() );

  // Load the titlebar center tile image (this will be used as
  //     the background for the caption bubble tiles).
  titleCenter = loadImage( "titlebar-center", QColor() );

  // Load the small version of the caption bubble corner & center images.
  captionRight  = loadImage( "caption-small-right",  captionColorActive );
  captionCenter = loadImage( "caption-small-center", captionColorActive );

  // Create the caption bubble tiles (by blending the images onto the titlebar)
  activeTiles[ CaptionSmallRight  ] = composite( captionRight,  titleCenter);
  activeTiles[ CaptionSmallCenter ] = composite( captionCenter, titleCenter );

  delete captionRight;
  delete captionCenter;

  // Create the titlebar center tile
  activeTiles[ TitleCenter ] = new QPixmap( *titleCenter );

  delete titleCenter;

  // Load the left & right border pixmaps
  activeTiles[ BorderLeft ]  = loadPixmap( "border-left",  QColor());
  activeTiles[ BorderRight ] = loadPixmap( "border-right",  QColor());

  activeTiles[ GrabBarLeft ] = loadPixmap( "bottom-left",   captionColorActive  );
  activeTiles[ GrabBarRight ] = loadPixmap( "bottom-right",  captionColorActive   );
  activeTiles[ GrabBarCenter ] = loadPixmap( "bottom-center", QColor());

  activeTiles[BorderBottomLeft]=loadPixmap("border-bottom-left",captionColorActive );
  activeTiles[BorderBottomRight]=loadPixmap("border-bottom-right",captionColorActive );
  activeTiles[LeftBorderTitle]=loadPixmap("left-border-title",captionColorActive );

  // Inactive tiles
  // -------------------------------------------------------------------------
  inactiveTiles[ TitleLeft ]  = loadPixmap( "titlebar-left-inactive", captionColorInactive);
  inactiveTiles[ TitleRight ] = loadPixmap( "titlebar-right-inactive", QColor());

  titleCenter = loadImage( "titlebar-center-inactive", QColor());

  captionRight  = loadImage( "caption-small-right-inactive",  captionColorInactive);
  captionCenter = loadImage( "caption-small-center-inactive", captionColorInactive);

  inactiveTiles[ CaptionSmallRight ]  = composite( captionRight,  titleCenter );
  inactiveTiles[ CaptionSmallCenter ] = composite( captionCenter, titleCenter );

  delete captionRight;
  delete captionCenter;

  inactiveTiles[ TitleCenter ] = new QPixmap( *titleCenter );

  delete titleCenter;

  inactiveTiles[ BorderLeft ]  = loadPixmap( "border-left",  QColor() );
  inactiveTiles[ BorderRight ] = loadPixmap( "border-right", QColor() );

  inactiveTiles[ GrabBarLeft ]   = loadPixmap( "bottom-left-inactive", captionColorInactive );
  inactiveTiles[ GrabBarRight ]  = loadPixmap( "bottom-right-inactive",  captionColorInactive);
  inactiveTiles[ GrabBarCenter ] = loadPixmap( "bottom-center", QColor());

  inactiveTiles[BorderBottomLeft]=loadPixmap("border-bottom-left-inactive",captionColorInactive);
  inactiveTiles[BorderBottomRight]=loadPixmap("border-bottom-right-inactive",captionColorInactive);
  inactiveTiles[LeftBorderTitle]=loadPixmap("left-border-title-inactive", captionColorInactive );


  // Buttons
  // -------------------------------------------------------------------------
  buttonColor  = QColor(); //KDecoration::options()->color( Options::ButtonBg, true );

  titleButtonRoundNormal  = loadPixmap( "titlebutton-round",  buttonColor );
  titleButtonRoundHover  = loadPixmap( "titlebutton-round",  QColor("#A5A6A5") );
  titleButtonRoundDown  = loadPixmap( "titlebutton-round",  QColor("#666666") );


  buttonDecos[ Sticky ]   = loadPixmap( "sticky-windows",  buttonColor );
  buttonDecos[ Unsticky ] = loadPixmap( "unsticky-windows",  buttonColor );
  buttonDecos[ Help ]     = loadPixmap( "help-windows",  buttonColor );
  buttonDecos[ Iconify ]  = loadPixmap( "iconify-windows",  buttonColor );
  buttonDecos[ Maximize ] = loadPixmap( "maximize-windows",  buttonColor );
  buttonDecos[ Restore ]  = loadPixmap( "restore-windows",  buttonColor );
  buttonDecos[ Close ]    = loadPixmap( "close-windows",  buttonColor  );
  if( QColor("#EDEDED") == KDecoration::options()->color( ColorTitleBar,   false ))
    captionColorInactive=QColor("#EDEDED");
  buttonDecos[ MaximizeInactive ] = loadPixmap( "maximize-windows",  captionColorInactive );
  buttonDecos[ RestoreInactive ]  = loadPixmap( "restore-windows",  captionColorInactive);
  buttonDecos[ CloseInactive ]    = loadPixmap( "close-windows",  captionColorInactive );
  buttonDecos[ IconifyInactive ]  = loadPixmap( "iconify-windows",  captionColorInactive );
  buttonDecos[ HelpInactive ]     = loadPixmap( "help-windows",  captionColorInactive );

  // Prepare the tiles for use
  // -------------------------------------------------------------------------
  if ( QApplication::reverseLayout() ) {

    flip( activeTiles[TitleLeft], activeTiles[TitleRight] );
    flip( inactiveTiles[TitleLeft], inactiveTiles[TitleRight] );

    flip( activeTiles[CaptionSmallRight]);
    flip( inactiveTiles[CaptionSmallRight]);

    flip( activeTiles[BorderLeft], activeTiles[BorderRight] );
    flip( inactiveTiles[BorderLeft], inactiveTiles[BorderRight] );

    flip( activeTiles[GrabBarLeft], activeTiles[GrabBarRight] );
    flip( inactiveTiles[GrabBarLeft], inactiveTiles[GrabBarRight] );

    flip( activeTiles[BorderBottomLeft], activeTiles[BorderBottomRight]);
    flip( inactiveTiles[BorderBottomLeft], inactiveTiles[BorderBottomRight]);
    flip( activeTiles[BorderBottomLeft], activeTiles[BorderBottomRight]);
    flip( inactiveTiles[BorderBottomLeft], inactiveTiles[BorderBottomRight]);

    flip( titleButtonRoundNormal );
    flip( titleButtonRoundHover );
    flip( titleButtonRoundDown );
  }

  // Pretile the center & border tiles for optimal performance
  pretile( activeTiles[ CaptionSmallCenter ], 64, Qt::Horizontal );
  pretile( activeTiles[ TitleCenter ], 128, Qt::Horizontal );
  pretile( activeTiles[ GrabBarCenter ], 128, Qt::Horizontal );
  pretile( activeTiles[ BorderLeft ], 128, Qt::Vertical );
  pretile( activeTiles[ BorderRight ], 128, Qt::Vertical );

  pretile( inactiveTiles[ CaptionSmallCenter ], 64, Qt::Horizontal );
  pretile( inactiveTiles[ TitleCenter ], 128, Qt::Horizontal );
  pretile( inactiveTiles[ GrabBarCenter ], 128, Qt::Horizontal );
  pretile( inactiveTiles[ BorderLeft ], 128, Qt::Vertical );
  pretile( inactiveTiles[ BorderRight ], 128, Qt::Vertical );
}

void MandrakeHandler::addWidth (int width, QPixmap *&pix, bool left, QPixmap *bottomPix) {
	int w = pix->width()+width;
	int h = pix->height();

	QPixmap *tmp = new QPixmap (w, h);
	tmp->fill ();
	QPainter p;
	p.begin (tmp);

	for (int i = 0; i < h; i++)
		p.drawPixmap (0, i, *bottomPix, i%2, 0, w,1);

	if (left)
		p.drawPixmap(0, 0, *pix);
	else
		p.drawPixmap(width, 0, *pix);

	p.end();

	delete pix;
	pix = tmp;
}


void MandrakeHandler::addHeight (int height, QPixmap *&pix) {
	int w = pix->width();
	int h = pix->height()+height;

	QPixmap *tmp = new QPixmap (w, h);
	QPainter p;
	p.begin (tmp);
	if (pix->height() > 10) {
		p.drawPixmap(0, 0, *pix, 0, 0, w, 11);
		for (int i = 0; i < height; i+=2)
			p.drawPixmap(0, 11+i, *pix, 0, 11, w, 2);
		p.drawPixmap(0, 11+height, *pix, 0, 11, w, -1);
	}
	else {
		int lines  = h-3;
		int factor = pix->height()-3;
		for (int i = 0; i < lines; i++)
			p.drawPixmap(0, i, *pix, 0, i*factor/lines, w, 1);
		p.drawPixmap(0, lines, *pix, 0, factor, w, 3);
	}
	p.end();

	delete pix;
	pix = tmp;
}


KDecoration* MandrakeHandler::createDecoration( KDecorationBridge* bridge )
{
        return new MandrakeClient( bridge, this );
}


void MandrakeHandler::destroyPixmaps()
{
    for ( int i = 0; i < NumTiles; i++ ) {
        if ( activeTiles[i] ) {
            delete activeTiles[i];
            activeTiles[i] = NULL;
        }

        if ( inactiveTiles[i] ) {
            delete inactiveTiles[i];
            inactiveTiles[i] = NULL;
        }
    }

    delete titleButtonRoundNormal;
    delete titleButtonRoundHover;
    delete titleButtonRoundDown;

}


void MandrakeHandler::flip( QPixmap *&pix1, QPixmap *&pix2 )
{
	// Flip the pixmaps horizontally
	QPixmap *tmp = new QPixmap( pix1->xForm( QWMatrix(-1,0,0,1,pix1->width(),0) ) );
	delete pix1;
	pix1 = new QPixmap( pix2->xForm( QWMatrix(-1,0,0,1,pix2->width(),0) ) );
	delete pix2;
	pix2 = tmp;
}


void MandrakeHandler::flip( QPixmap *&pix )
{
	// Flip the pixmap horizontally
	QPixmap *tmp = new QPixmap( pix->xForm( QWMatrix(-1,0,0,1,pix->width(),0) ) );
	delete pix;
	pix = tmp;
}


void MandrakeHandler::pretile( QPixmap *&pix, int size, Qt::Orientation dir )
{
    QPixmap *newpix;
    QPainter p;

    if ( dir == Qt::Horizontal )
        newpix = new QPixmap( size, pix->height() );
    else
        newpix = new QPixmap( pix->width(), size );

    p.begin( newpix );
    p.drawTiledPixmap( newpix->rect(), *pix ) ;
    p.end();

    delete pix;
    pix = newpix;
}


void MandrakeHandler::readConfig()
{
    KConfig *c = new KConfig( "kwinmandrakerc" );

    c->setGroup( "General" );
    shadowedText = c->readBoolEntry( "UseShadowedText", true );
    showIcons = c->readBoolEntry( "ShowAppIcons", true );
    if ( ! settings_cache ) {
        settings_cache = new SettingsCache;

        if ( KDecoration::options()->customButtonPositions() ) {
            settings_cache->buttonsLeft  = KDecoration::options()->titleButtonsLeft();
            settings_cache->buttonsRight = KDecoration::options()->titleButtonsRight();
        } else {
            settings_cache->buttonsLeft  = QString( default_left );
            settings_cache->buttonsRight = QString( default_right );
        }

        settings_cache->aTitleColor   = KDecoration::options()->color( ColorTitleBar,   true );
        settings_cache->aTitleBlend   = KDecoration::options()->color( ColorTitleBlend, true );
        settings_cache->iTitleColor   = KDecoration::options()->color( ColorTitleBar,   false );
        settings_cache->iTitleBlend   = KDecoration::options()->color( ColorTitleBlend, false );
        settings_cache->buttonColor   = QColor();//KDecoration::options()->color( Options::ButtonBg,   true );
        settings_cache->showTooltips  = KDecoration::options()->showTooltips();
    }

    delete c;
}


QPixmap *MandrakeHandler::composite( QImage *over, QImage *under )
{
    QImage dest( over->width(), over->height(), 32 );
    int width = over->width(), height = over->height();

    // Clear the destination image
    Q_UINT32 *data = reinterpret_cast<Q_UINT32*>( dest.bits() );
    for (int i = 0; i < width * height; i++)
        *(data++) = 0;
    // Copy the under image (bottom aligned) to the destination image
    for (int y1 = height - under->height(), y2 = 0; y1 < height; y1++, y2++ )
    {
        register Q_UINT32 *dst = reinterpret_cast<Q_UINT32*>( dest.scanLine(y1) );
        register Q_UINT32 *src = reinterpret_cast<Q_UINT32*>( under->scanLine(y2) );

        //juste one pixel !!!!
        for ( int x = 0; x < width; x++ )
            *(dst++) = *(src);
    }
    // Blend the over image onto the destination
    register Q_UINT32 *dst = reinterpret_cast<Q_UINT32*>( dest.bits() );
    register Q_UINT32 *src = reinterpret_cast<Q_UINT32*>( over->bits() );
    for ( int i = 0; i < width * height; i++ )
    {
        int r1 = qRed( *dst ), g1 = qGreen( *dst ), b1 = qBlue( *dst );
        int r2 = qRed( *src ), g2 = qGreen( *src ), b2 = qBlue( *src );
        int a  = qAlpha( *src );

        if ( a == 0xff )
            *dst = *src;

        else if ( a != 0x00 )
            *dst = qRgba( Q_UINT8( r1 + (((r2 - r1) * a) >> 8) ),
                          Q_UINT8( g1 + (((g2 - g1) * a) >> 8) ),
                          Q_UINT8( b1 + (((b2 - b1) * a) >> 8) ),
                          0xff );

        else if ( qAlpha(*dst) == 0x00 )
            *dst = 0;

        src++; dst++;
    }

    // Create the final pixmap and return it
    return new QPixmap( dest );
}


QImage *MandrakeHandler::loadImage( const QString &name, const QColor &col )
{
    if ( col.isValid() )
    {
        QImage *img = new QImage( imageDb->image(name)->copy() );
        KIconEffect::colorize( *img, col, 1.0 );
        return img;
    }
    else
        return new QImage( imageDb->image(name)->copy() );
}


QPixmap *MandrakeHandler::loadPixmap( const QString &name, const QColor &col )
{
	QImage *img = loadImage( name, col );
	QPixmap *pix = new QPixmap( *img );
	delete img;
	return pix;
}


bool MandrakeHandler::reset(unsigned long changed)
{
  QString buttonsLeft, buttonsRight;

  mandrake_initialized = false;

  bool needHardReset  = false;
  bool pixmapsInvalid = false;

  // Re-read the config file
  readConfig();

  // Check if the color scheme has changed
  if ( changed & SettingColors )
    {
      pixmapsInvalid = true;
      needHardReset  = true;
    }
  // Check if button positions have changed

  if ( changed & SettingButtons ) {
    needHardReset = true;

  }

  // Check if tooltips options have changed
  if ( changed & SettingTooltips ) {
    needHardReset = true;
  }
  // Do we need to recreate the pixmaps?
  if ( pixmapsInvalid ) {
    destroyPixmaps();
    createPixmaps();
  }

  mandrake_initialized = true;

  // Do we need to "hit the wooden hammer" ?
  if ( needHardReset )
    resetDecorations( changed );
  return needHardReset;

}



const QPixmap *MandrakeHandler::tile( TilePixmap tilePix, bool active ) const
{
	return ( active ? activeTiles[ tilePix ] : inactiveTiles[ tilePix ] );
}

// -------------------------------------------------------------------------------------------


MandrakeMenuButton::MandrakeMenuButton(MandrakeClient* parent, const char *name, Button btn, const QString &tip, bool _captionButton)
    : MandrakeButton(parent, name, btn, tip, _captionButton )
{
    activeIcon=NULL;
    inactiveIcon=NULL;
}


MandrakeMenuButton::~MandrakeMenuButton()
{
    delete activeIcon;
    delete inactiveIcon;
}

void MandrakeMenuButton::enterEvent( QEvent *e )
{
    MandrakeButton::enterEvent( e );
}

void MandrakeMenuButton::leaveEvent( QEvent *e )
{
    MandrakeButton::leaveEvent( e );
}

void MandrakeMenuButton::mousePressEvent( QMouseEvent *e )
{
    MandrakeButton::mousePressEvent( e );
}


void MandrakeMenuButton::mouseReleaseEvent( QMouseEvent *e )
{
    MandrakeButton::mouseReleaseEvent( e );
}

QPixmap *MandrakeMenuButton::menuIcons()
{
    QPixmap *icon;
    //FIXME !!!!!!!
    if ( client->isActive()) {
        if ( !activeIcon )
	  activeIcon = new QPixmap(client->icon().pixmap( QIconSet::Small, QIconSet::Normal));
        icon = activeIcon;
    } else {
        if ( !inactiveIcon ) {
	  QImage img = client->icon().pixmap( QIconSet::Small, QIconSet::Normal ).convertToImage();
            KIconEffect::semiTransparent( img );
            inactiveIcon = new QPixmap( img );
        }
        icon = inactiveIcon;
    }
    return icon;
}


void MandrakeMenuButton::drawBackgroundButton( QPainter *p )
{
    // Draw the button background
    if( captionButton )
        p->drawPixmap( 0, 0, *clientHandler->tile( CaptionSmallCenter, client->isActive() ),
                       0, 3, 19, 19 );
    else
        p->drawPixmap( 0, 0, *clientHandler->tile( TitleCenter, client->isActive() ),
                       0, 3, 19, 19 );
#if 0 //FIXME
    if ( isDown() )
        deco = clientHandler->roundButtonDown();
    else if ( hover )
        deco = clientHandler->roundButtonHover();
    else
        deco = clientHandler->roundButtonNormal();
#endif
    p->drawPixmap( QPoint(), *menuIcons(), QStyle::visualRect( QRect(0, 0, 19, 19), menuIcons()->rect() ) );

}

void MandrakeMenuButton::drawButton( QPainter *p )
{
    drawBackgroundButton( p );
}

void MandrakeMenuButton::iconChange()
{
    // Force updateCaptionBuffer() to recreate the cached icons
    delete activeIcon;
    delete inactiveIcon;
    activeIcon = inactiveIcon = NULL;
    (void)menuIcons();
    repaint( false );
}


MandrakeButton::MandrakeButton( MandrakeClient* parent, const char *name, Button btn, const QString &tip, bool _captionButton )
    : QButton( parent->widget(), name ),
      client( parent ), button( btn ), hover( false ), lastbutton( 0 )
{
  QToolTip::add( this, tip ); // FRAME

  captionButton=_captionButton;
  setCursor( arrowCursor );

  setBackgroundMode( NoBackground );
  setFixedSize( 19, 19 );

  bool sticky = (button == OnAllDesktopsButton);
  setToggleButton( sticky );
}


MandrakeButton::~MandrakeButton()
{
	// Empty.
}


void MandrakeButton::enterEvent( QEvent *e )
{
	QButton::enterEvent( e );

	hover = true;
	repaint( false );
}


void MandrakeButton::leaveEvent( QEvent *e )
{
	QButton::leaveEvent( e );

	hover = false;
	repaint( false );
}


void MandrakeButton::mousePressEvent( QMouseEvent *e )
{
	lastbutton = e->button();
	QMouseEvent me( e->type(), e->pos(), e->globalPos(), LeftButton, e->state() );
	QButton::mousePressEvent( &me );
	//	if(lastbutton == StickyButton)
	//repaint();

}


void MandrakeButton::mouseReleaseEvent( QMouseEvent *e )
{
	lastbutton = e->button();
	QMouseEvent me( e->type(), e->pos(), e->globalPos(), LeftButton, e->state() );
	QButton::mouseReleaseEvent( &me );
}


void MandrakeButton::drawBackgroundButton( QPainter *p, bool drawDeco )
{
    // Draw the button background
    if( captionButton )
        p->drawPixmap( 0, 0, *clientHandler->tile( CaptionSmallCenter, client->isActive() ),
                       0, 3, 19, 19 );
    else
        p->drawPixmap( 0, 0, *clientHandler->tile( TitleCenter, client->isActive() ),
                       0, 3, 19, 19 );
    if( drawDeco )
    {
        const QPixmap *deco;
        if ( isDown() )
            deco = clientHandler->roundButtonDown();
        else if ( hover )
            deco = clientHandler->roundButtonHover();
        else
            deco = clientHandler->roundButtonNormal();

        p->drawPixmap( QPoint(), *deco, QStyle::visualRect( QRect(0, 0, 19, 19), deco->rect() ) );
    }

}

void MandrakeButton::drawButton( QPainter *p )
{
    const QPixmap *deco;
    /*CloseInactive, RestoreInactive, MaximizeInactive*/
    // Draw the button deco on the bevel
    switch ( button ) {
    case OnAllDesktopsButton:
        drawBackgroundButton( p,false );
        deco = clientHandler->buttonDeco( isOn() ? Unsticky : Sticky );
        break;

    case HelpButton:
        drawBackgroundButton( p );
        deco = clientHandler->buttonDeco( (client->isActive()|| hover) ? Help : HelpInactive );
        // The '?' won't be flipped around in the ctor, so we need to
        //  shift it to the right to compensate for the button shadow
        //  being on the left side of the button in RTL mode.
        if ( QApplication::reverseLayout() )
            p->translate( 2, 0 );
        break;

    case MinButton:
        drawBackgroundButton( p );
        deco = clientHandler->buttonDeco( (client->isActive()|| hover) ? Iconify :IconifyInactive );
        break;

    case MaxButton:
        drawBackgroundButton( p );
	if( client->isActive()|| hover )
	  deco = clientHandler->buttonDeco( client->maximizeMode() == MandrakeClient::MaximizeFull ? Restore : Maximize );
	else
	  deco = clientHandler->buttonDeco( client->maximizeMode() == MandrakeClient::MaximizeFull ? RestoreInactive : MaximizeInactive );

        break;

    case CloseButton:
        drawBackgroundButton( p );
        deco = clientHandler->buttonDeco( (client->isActive()|| hover) ? Close : CloseInactive );
        break;

    default:
        deco = NULL;
    }
    if( deco)
      p->drawPixmap( QPoint(), *deco, QStyle::visualRect( QRect(0, 0, 19, 19), deco->rect() ) );
    //p->setPen( Qt::red ); // debug
    //p->drawRect( QStyle::visualRect( QRect(0, 0, 18, 18), deco->rect()) );    // debug

}



// ------------------------------------------------------------------------------------------



MandrakeClient::MandrakeClient( KDecorationBridge* bridge, KDecorationFactory* factory )
  : KDecoration( bridge, factory ),
      captionBufferDirty( true ), maskDirty( true )
{
}


MandrakeClient::~MandrakeClient()
{
}

void MandrakeClient::init()
{
  createMainWidget( WStaticContents | WResizeNoErase | WRepaintNoErase );
  widget()->installEventFilter( this );

    // Minimize flicker
    widget()->setBackgroundMode( NoBackground );

    for ( int i=0; i < NumButtons; i++ )
        button[i] = NULL;
    menuButton=NULL;

    createLayout();
}

void MandrakeClient::createLayout()
{
    QVBoxLayout *mainLayout   = new QVBoxLayout( widget() );
    QBoxLayout *titleLayout   = new QBoxLayout( 0, QBoxLayout::LeftToRight, 0, 0, 0 );
    QHBoxLayout *windowLayout = new QHBoxLayout();


    int grabBarHeight = clientHandler->grabBarHeight();
    int topSpacing = 0;//todo change it.
    QSpacerItem *topSpacer = new QSpacerItem( 10, topSpacing,
                               QSizePolicy::Expanding, QSizePolicy::Minimum );

    mainLayout->addItem( topSpacer );

    mainLayout->addLayout( titleLayout );     // Titlebar
    mainLayout->addLayout( windowLayout, 10 ); // Left border + window + right border
    mainLayout->addSpacing( grabBarHeight+1 );  // Bottom grab bar

    titleLayout->setSpacing( buttonSpacing );
    titleLayout->addSpacing( buttonMargin );      // Left button margin
    addButtons( titleLayout, options()->customButtonPositions() ?
                options()->titleButtonsLeft() : QString(default_left), /*QApplication::reverseLayout() ? false : true*/true );

    titlebar = new QSpacerItem( 10, clientHandler->titleBarHeight()
                                /*- topSpacing*/, QSizePolicy::Expanding, QSizePolicy::Minimum );
    titleLayout->addItem( titlebar );
    titleLayout->addSpacing( buttonSpacing );

    addButtons( titleLayout, options()->customButtonPositions() ?
                options()->titleButtonsRight() : QString(default_right), /*QApplication::reverseLayout() ? true : false*/false );

    titleLayout->addSpacing( buttonMargin - 1 );  // Right button margin

    windowLayout->addSpacing( 5 );                // Left border
    if( isPreview())
      windowLayout->addWidget( new QLabel( i18n( "<center><b>Galaxy</b></center>" ), widget()));   // Window wrapper FRAME
    else
      windowLayout->addItem( new QSpacerItem( 0, 0 )); //no widget in the middle

    windowLayout->addSpacing( 5 );                // Right border

}


void MandrakeClient::reset( unsigned long )
{

    calculateCaptionRect();

    captionBufferDirty = maskDirty = true;

    // Only repaint the window if it's visible
    // (i.e. not minimized and on the current desktop)
    if ( widget()->isVisible() ) {
        widget()->repaint( false );

        for ( int i = 0; i < NumButtons; i++ )
            if ( button[i] ) button[i]->repaint( false );
    }

}


void MandrakeClient::addButtons( QBoxLayout *layout, const QString &s, bool _captionButton )
{
    for ( uint i=0; i < s.length(); i++ )
    {
        bool addSpacing = false;
        switch ( s[i].latin1() )
        {
            // Menu button
        case 'M' :
            if ( !menuButton ) {
                menuButton = new MandrakeMenuButton( this, "menu", MenuButton, i18n("Menu"),_captionButton );
                connect( menuButton, SIGNAL( pressed() ), SLOT( menuButtonPressed() ) );
                layout->addWidget( menuButton );
                addSpacing = true;
            }
            break;

            // Sticky button
        case 'S' :
            if ( !button[OnAllDesktopsButton] ) {
                button[OnAllDesktopsButton] = new MandrakeButton( this, "sticky", OnAllDesktopsButton, i18n("that means, stick to all desktops","Sticky"),_captionButton );
                connect( button[OnAllDesktopsButton], SIGNAL( clicked() ), SLOT( toggleOnAllDesktops() ) );
                layout->addWidget( button[OnAllDesktopsButton] );
                addSpacing = true;
            }
            break;

            // Help button
        case 'H' :
            if ( !button[HelpButton] && providesContextHelp() ) {
                button[HelpButton] = new MandrakeButton( this, "help", HelpButton, i18n("Help"),_captionButton );
                connect( button[HelpButton], SIGNAL( clicked() ), SLOT( showContextHelp() ) );
                layout->addWidget( button[HelpButton] );
                addSpacing = true;
            }
            break;

            // Minimize button
        case 'I' :
            if ( !button[MinButton] && isMinimizable() ) {
                button[MinButton] = new MandrakeButton( this, "iconify", MinButton, i18n("Minimize"),_captionButton );
                connect( button[MinButton], SIGNAL( clicked() ), SLOT( minimize() ) );
                layout->addWidget( button[MinButton] );
                addSpacing = true;
            }
            break;

            // Maximize button
        case 'A' :
            if ( !button[MaxButton] && isMaximizable() ) {
                button[MaxButton] = new MandrakeButton( this, "maximize", MaxButton, i18n("Maximize"), _captionButton );
                connect( button[MaxButton], SIGNAL( clicked() ), SLOT( slotMaximize() ) );
                layout->addWidget( button[MaxButton] );
                addSpacing = true;
            }
            break;

            // Close button
        case 'X' :
            if ( !button[CloseButton] && isCloseable() ) {
                button[CloseButton] = new MandrakeButton( this, "close", CloseButton, i18n("Close"), _captionButton );
                connect( button[CloseButton], SIGNAL( clicked() ), SLOT( closeWindow() ) );
                layout->addWidget( button[CloseButton] );
                addSpacing = true;
            }
            break;

            // Additional spacing
        case '_' :
            layout->addSpacing( buttonSpacing );
            addSpacing = true;
            break;
        }
        if( addSpacing)
            layout->addSpacing( buttonSpacing );
    }
}


void MandrakeClient::updateMask()
{
    if ( !mandrake_initialized )
        return;
    // To maximize performance this code uses precalculated bounding rects
    // to set the window mask. This saves us from having to allocate a 1bpp
    // pixmap, paint the mask on it and then have the X server iterate
    // over the pixels to compute the bounding rects from it.

    // To elliminate any possible overhead introduced by Qt it uses Xlib
    // calls directly, which also has the advantage that it allows us to
    // optimize the sorting of the rectangles.

    QRegion r;
    register int w, y = 0;
    int nrects;

    nrects = 7;

    w = width();
    r += QRegion( 3, y++, w-6 ,  1 );
    r += QRegion( 2, y++, w-4 ,  1 );
    r += QRegion( 1, y++, w-2,  1 );
    r += QRegion( 1, height() -3, w-2, 1);
    r += QRegion( 2, height() -2, w-4, 1);
    r += QRegion( 3, height() -1, w-6, 1);
    // The part of the window below the titlebar
    r += QRegion( 0, y, w, height() - y - 3 );

    setMask( r, YXBanded );
    maskDirty = false;
}


void MandrakeClient::updateCaptionBuffer()
{
  if ( !mandrake_initialized )
    return;

  bool active = isActive();
  if ( captionBuffer.size() != captionRect.size() )
    captionBuffer.resize( captionRect.size() );

  if ( captionBuffer.isNull() )
    return;


  QPainter p( &captionBuffer );

  if( QApplication::reverseLayout() )
    {
      p.drawPixmap( 0, 0, *clientHandler->tile( CaptionSmallRight, active ) );

      p.drawTiledPixmap( 28, 0 , captionRect.width()-23, captionRect.height(),
			 *clientHandler->tile( CaptionSmallCenter, active ) );
    }
  else
    {
      p.drawTiledPixmap( 0, 0, captionRect.width() - 33, captionRect.height(),
			 *clientHandler->tile( CaptionSmallCenter, active ) );
      p.drawPixmap( captionRect.width() - 33, 0, *clientHandler->tile( CaptionSmallRight, active ) );
    }
#if 0
  if ( clientHandler->showAppIcons() )
    {
      if ( active ) {
	if ( ! activeIcon )
	  activeIcon = new QPixmap( this->icon().pixmap( QIconSet::Small, QIconSet::Normal )); // FRAME
	icon = activeIcon;
      } else {
	if ( ! inactiveIcon ) {
	  QImage img = this->icon().pixmap( QIconSet::Small, QIconSet::Normal ).convertToImage();
	  KIconEffect::semiTransparent( img );
	  inactiveIcon = new QPixmap( img );
	}
	icon = inactiveIcon;
      }
    }
#endif
  p.setFont( options()->font( active ) );

  QString tmp = options()->customButtonPositions() ?
    options()->titleButtonsLeft() : QString(default_left);

  int len=tmp.length();
  if( len )
    len= calculateLeftButtonWidth(tmp);
  int addwidth =0;
  if( QApplication::reverseLayout())
    {
      tmp = options()->customButtonPositions() ?
	options()->titleButtonsRight() : QString(default_right);
      addwidth = tmp.length();
      if( addwidth )
	addwidth= calculateLeftButtonWidth(tmp);

    }

  int xpos = /*QMAX( (captionRect.width() - tw) / 3, 8 )+*/len+8;
  QRect tr = QStyle::visualRect( QRect(xpos+addwidth, 1, captionRect.width() - xpos - 20 + addwidth,
				       captionRect.height() - 4), captionBuffer.rect() );

  //p.setPen( Qt::red ); // debug
  //p.drawRect( tr );    // debug
#if 0
  // Application icon
  if ( clientHandler->showAppIcons() )
    {
      QRect iconRect = QStyle::visualRect( QRect(tr.x(),
						 1 + (captionRect.height() - 4 - 16) / 2, 16, 16), tr );
      QRect r( icon->rect() );
      r.moveCenter( iconRect.center() );

      if ( tr.width() > 16 ) {
	p.drawPixmap( r, *icon );
      } else {
	QRect sr( 0, 0, icon->width(), icon->height() );

	if ( QApplication::reverseLayout() )
	  sr.addCoords( icon->width() - tr.width(), 0, 0, 0 );
	else
	  sr.addCoords( 0, 0, -( icon->width() - tr.width() ), 0 );

	p.drawPixmap( r.x() + sr.x(), r.y() + sr.y(), *icon,
		      sr.x(), sr.y(), sr.width(), sr.height() );
      }

      //p.drawRect( r ); // debug

      if ( QApplication::reverseLayout() )
	tr.addCoords( 0, 0, -(16 + iconSpacing), 0 );
      else
	tr.addCoords( (16 + iconSpacing), 0, 0, 0 );
    }
#endif
  // Draw the titlebar text
  int flags = AlignVCenter | SingleLine;
  flags |= ( QApplication::reverseLayout() ? AlignRight : AlignLeft );

  if ( clientHandler->useShadowedText() )
    {
      p.translate( QApplication::reverseLayout() ? -1 : 1, 1 );
      p.setPen( options()->color(ColorTitleBar, active).dark() );
      p.setPen( black );
      p.drawText( tr, flags, caption() );
      p.translate( QApplication::reverseLayout() ? 1 : -1, -1 );
    }

  p.setPen( options()->color( ColorFont, active ) );
  p.drawText( tr, flags, caption() );


  captionBufferDirty = false;
}

void MandrakeClient::calculateCaptionRect()
{
    QFontMetrics fm( options()->font(isActive()) );
    int cw = fm.width( caption() ) + /*95*/45;

    QString tmp = options()->customButtonPositions() ?
                  options()->titleButtonsLeft() : QString(default_left);

    int len=tmp.length();
    int leftButtonWidth =0;
    if( len )
        leftButtonWidth= calculateLeftButtonWidth(tmp);
    cw+=leftButtonWidth;
    cw = QMIN( cw, titlebar->geometry().width() );
    if( cw <leftButtonWidth+15)
        cw=leftButtonWidth+15;

    int addwidth =0;
    if( QApplication::reverseLayout())
      {
	tmp = options()->customButtonPositions() ?
	  options()->titleButtonsRight() : QString(default_right);
	addwidth = tmp.length();
	if( addwidth )
	  addwidth= calculateLeftButtonWidth(tmp);
    }


    //minimum titlebar size
    if ( cw < 77 )
        cw = 77;
    captionRect = QStyle::visualRect( QRect(/*titlebar->geometry().x()*/0, 0,
                                            cw+addwidth, clientHandler->titleBarHeight() ),
                                      titlebar->geometry() );
}

int MandrakeClient::calculateLeftButtonWidth(const QString &s)const
{
    int width=0;
    for ( uint i=0; i < s.length(); i++ )
    {
        bool addSpacing = false;
        switch ( s[i].latin1() )
        {
            // Menu button
        case 'M' :
        case 'S' :
            width+=19;
            addSpacing = true;
            break;
        case 'H' :
            if( providesContextHelp())
            {
                width+=19;
                addSpacing = true;
            }
            break;
            // Minimize button
        case 'I' :
            if ( isMinimizable() ) {
                addSpacing = true;
                width+=19;
            }
            break;
            // Maximize button
        case 'A' :
            if ( isMaximizable() ) {
                width+=19;
                addSpacing = true;
            }
            break;
            // Close button
        case 'X' :
            if ( !button[CloseButton] && isCloseable() ) {
                width+=19;
                addSpacing = true;
            }
            break;
            // Additional spacing
        case '_' :
            addSpacing = true;
            break;
        }
        if( addSpacing)
        {
            width+=iconSpacing;
        }
    }
    return width;
}


void MandrakeClient::captionChange()
{
	QRect r( captionRect );
	calculateCaptionRect();

	if ( r.size() != captionRect.size() )
		maskDirty = true;

	captionBufferDirty = true;

	widget()->repaint( r | captionRect, false );
}


void MandrakeClient::iconChange()
{
#if 0
	if ( clientHandler->showAppIcons() ) {

		// Force updateCaptionBuffer() to recreate the cached icons
		if ( activeIcon )
			delete activeIcon;

		if ( inactiveIcon )
			delete inactiveIcon;

		activeIcon = inactiveIcon = NULL;

		captionBufferDirty = true;
		widget()->repaint( captionRect, false );
	}

#endif
    if( menuButton )
    {
        menuButton->iconChange();
        captionBufferDirty = true;
        widget()->repaint( captionRect, false );
    }
}


void MandrakeClient::activeChange( )
{
    captionBufferDirty = true;

    widget()->repaint( false );

    for ( int i=0; i < NumButtons; i++ )
        if ( button[i] ) button[i]->repaint( false );
    iconChange();
}


void MandrakeClient::maximizeChange()
{
    if ( button[ MaxButton ] ) {
      QToolTip::remove( button[ MaxButton ] );
      QToolTip::add( button[ MaxButton ], maximizeMode() == MaximizeFull ? i18n("Restore") : i18n("Maximize") );
        button[ MaxButton ]->repaint();
    }
}

void MandrakeClient::desktopChange()
{
  if ( button[ OnAllDesktopsButton ] )
    {
      QToolTip::remove( button[ OnAllDesktopsButton ] );
      QToolTip::add( button[ OnAllDesktopsButton ], isOnAllDesktops() ? i18n("Not On All Desktops") : i18n("On All Desktops") );
    }
}


void MandrakeClient::menuButtonPressed()
{
	QPoint menuPoint ( menuButton->rect().bottomLeft().x() - 6,
					   menuButton->rect().bottomLeft().y() + 3 );
	KDecorationFactory* f = factory();
	showWindowMenu( menuButton->mapToGlobal( menuPoint ));
	if( !f->exists( this )) // 'this' was destroyed
	  return;
	menuButton->setDown(false);
}


void MandrakeClient::slotMaximize()
{
  switch ( button[ MaxButton ]->lastButton() )
    {
    case MidButton:
      maximize( maximizeMode() ^ MaximizeVertical );
      break;

    case RightButton:
      maximize( maximizeMode() ^ MaximizeHorizontal );
      break;

    case LeftButton:
      maximize( maximizeMode() == MaximizeFull ? MaximizeRestore : MaximizeFull );
      break;
    }
}


void MandrakeClient::paintEvent( QPaintEvent *e )
{
    if ( !mandrake_initialized )
        return;

    QPainter p( widget() );
    QRect updateRect( e->rect() );
    bool active = isActive();

    int titleBaseY         = 0;
    int titleBarHeight     = clientHandler->titleBarHeight();
    int grabBarHeight      = clientHandler->grabBarHeight();
    int leftBorderWidth    = clientHandler->tile( BorderLeft, active )->width();
    int rightBorderWidth   = clientHandler->tile( BorderRight, active )->width();
    if ( maskDirty )
        updateMask();
    // Titlebar
    // -----------------------------------------------------------------------
    if ( updateRect.y() < titleBarHeight )
    {
        int titleBarBaseHeight = titleBarHeight - titleBaseY;

        if ( captionBufferDirty )
            updateCaptionBuffer();

        // Top left corner
        if ( updateRect.x() < 5 )
        {
            p.drawPixmap( 0, active ? titleBaseY-1 : titleBaseY,
                          *clientHandler->tile( TitleLeft, active ) );
        }
        // Space between the top left corner and the caption bubble
        if ( updateRect.x() < captionRect.left() && updateRect.right() >= 5 ) {
            int x1 = QMAX( 5, updateRect.x() );
            int x2 = QMIN( captionRect.left(), updateRect.right() );

            p.drawTiledPixmap( x1, titleBaseY, x2 - x1 + 1 +5, titleBarBaseHeight,
                               *clientHandler->tile( TitleCenter, active ) );
        }
        // Caption bubble
        if ( updateRect.x() <= captionRect.right() && updateRect.right() > 5 )
        {
            if ( captionRect.width() >= 5 )
                p.drawPixmap( 5 + captionRect.left(), active ? 0 : titleBaseY, captionBuffer );
            else
                p.drawTiledPixmap( 5 + captionRect.x(), titleBaseY, captionRect.width()-5,
                                   titleBarBaseHeight, *clientHandler->tile( TitleCenter, active ) );
        }
        // Space between the caption bubble and the top right corner
        if ( updateRect.right() > captionRect.right() && updateRect.x() < width() - 5 )
        {
            int x1 = QMAX( captionRect.right() + 1, updateRect.x() );
            int x2 = QMIN( width() - 5, updateRect.right() );

            p.drawTiledPixmap( x1, titleBaseY, x2 - x1 + 1, titleBarBaseHeight,
                               *clientHandler->tile(  QApplication::reverseLayout() ? CaptionSmallCenter : TitleCenter, active ) );
        }

        // Top right corner
        if ( updateRect.right() >= width() - 5 )
        {
            p.drawPixmap( width() - 5, titleBaseY,
                          *clientHandler->tile( QApplication::reverseLayout() ? CaptionSmallCenter : TitleRight, active ) );
        }
    }

    // Borders
    // -----------------------------------------------------------------------
    if ( updateRect.bottom() >= titleBarHeight &&
         updateRect.top() < height()-1 /*- grabBarHeight*/ )
    {
        int top    = QMAX( titleBarHeight, updateRect.top() );
        int bottom = QMIN( updateRect.bottom(), height() - grabBarHeight );

        // Left border
        if ( updateRect.x() < leftBorderWidth )
        {
            p.drawTiledPixmap( 0, top, leftBorderWidth, bottom - top +1,
                               *clientHandler->tile( BorderLeft, active ) );
            if( !isShade() )
                p.drawTiledPixmap( 0, height()-19,
                                   leftBorderWidth,19,
                                   *clientHandler->tile( BorderBottomLeft, active ) );
	    if( !QApplication::reverseLayout() )
	      {
		if( top < titleBarHeight +17 )
		  {
		    p.drawTiledPixmap( 0, top, leftBorderWidth, 17,
				       *clientHandler->tile( LeftBorderTitle, active ) );
		  }
	      }
        }

        // Right border
        if ( e->rect().right() > width() - rightBorderWidth - 1 )
        {
            p.drawTiledPixmap( width() - rightBorderWidth, top, rightBorderWidth,
                               bottom - top + 1, *clientHandler->tile( BorderRight, active ) );
            if( !isShade() )
                p.drawTiledPixmap( width() - rightBorderWidth, height()-19,
                                   rightBorderWidth,19 ,
                                   *clientHandler->tile( BorderBottomRight, active ) );
	    if( QApplication::reverseLayout() )
	      {
		if( top < titleBarHeight +17 )
		  {
		    p.drawTiledPixmap( width() - rightBorderWidth, top, leftBorderWidth, 17,
				       *clientHandler->tile( LeftBorderTitle, active ) );
		  }
	      }
        }
    }

    // Bottom grab bar
    // -----------------------------------------------------------------------
    if ( updateRect.bottom() >= height() - grabBarHeight ) {
        // Bottom left corner
        if ( updateRect.x() < 18 )
        {
            if( !isShade() )
                p.drawPixmap( 5, height() - grabBarHeight,
                              *clientHandler->tile( GrabBarLeft, active ) );
            else
                p.drawTiledPixmap( 0, height() - grabBarHeight, 23, grabBarHeight, *clientHandler->tile( GrabBarRight, active ) );
        }
        // Space between the left corner and the right corner
        if ( updateRect.x() < width() - 18 ) {
            int x1 = QMAX( 18, updateRect.x() );
            int x2 = QMIN( width() - 18, updateRect.right() );

            p.drawTiledPixmap( x1, height() - grabBarHeight, x2 - x1 + 1,
                               grabBarHeight, *clientHandler->tile( GrabBarCenter, active ) );
        }
        // Bottom right corner
        if ( updateRect.right() > width() - 18 )
        {
            if( !isShade() )
                p.drawPixmap( width() - 18, height() - grabBarHeight,
                              *clientHandler->tile( GrabBarRight, active ) );
            else
                p.drawTiledPixmap( width() - 18, height() - grabBarHeight, width(), grabBarHeight, *clientHandler->tile( GrabBarRight, active ) );
        }
    }
    // Extra drawline for the 1 pixel empty space QLayout leaves when a window is shaded.
    p.setPen( active ? Qt::black : QColor("#848684") );

    p.drawLine( leftBorderWidth, height() - grabBarHeight - 1,
                width() - rightBorderWidth - 1, height() - grabBarHeight - 1 );

    p.drawLine( leftBorderWidth, titleBarHeight - 1,
                width() - rightBorderWidth - 1, titleBarHeight - 1 );

    p.drawLine( leftBorderWidth-1, titleBarHeight - 1,
                leftBorderWidth-1, height()  - grabBarHeight - 1 );

    p.drawLine( width() - rightBorderWidth, titleBarHeight - 1,
                width() - rightBorderWidth, height() - grabBarHeight - 1 );

}

void MandrakeClient::mouseDoubleClickEvent( QMouseEvent *e )
{
  if ( QRect( 0, 0, width(), clientHandler->titleBarHeight( ) ).contains( e->pos() ) )
    titlebarDblClickOperation();
}

void MandrakeClient::resizeEvent( QResizeEvent *e )
{
 // FRAME   Client::resizeEvent( e );

    QRect r( captionRect );
    calculateCaptionRect();

    if ( r.size() != captionRect.size() )
        captionBufferDirty = true;

    maskDirty = true;

    if ( widget()->isVisible() )
    {
        widget()->update( widget()->rect() );
        int dx = 0;
        int dy = 0;

        if ( e->oldSize().width() != width() )
            dx = 32 + QABS( e->oldSize().width() -  width() );

        if ( e->oldSize().height() != height() )
            dy = 8 + QABS( e->oldSize().height() -  height() );

        if ( dy )
            widget()->update( 0, height() - dy, width(), dy );

        if ( dx )
        {
            widget()->update( width() - dx + 1, 0, dx, height() );
            widget()->update( QRect( QPoint(4,4), titlebar->geometry().bottomLeft() - QPoint(1,0) ) );
            widget()->update( QRect( titlebar->geometry().topRight(), QPoint( width() - 4,
                                                                    titlebar->geometry().bottom() ) ) );
            // Titlebar needs no paint event
            QApplication::postEvent( this, new QPaintEvent( titlebar->geometry(), FALSE ) );
        }
    }
}

MandrakeClient::Position MandrakeClient::mousePosition( const QPoint &p ) const
{
    if( !isResizable())
	return PositionCenter;
    int titleBaseY = 0;

    int leftBorder   = clientHandler->tile( BorderLeft, true )->width();
    int rightBorder  = width() - clientHandler->tile( BorderRight, true )->width() - 1;
    int bottomBorder = height() - clientHandler->grabBarHeight() - 1;

    // Test if the mouse is over the titlebar area
    if ( p.y() < titleBaseY + 11 ) {
        // Test for the top left corner
        if ( p.x() < leftBorder + 11 ) {
            if ( (p.y() < titleBaseY + 3 && p.x() < leftBorder + 11) ||
                 (p.y() < titleBaseY + 6 && p.x() < leftBorder + 6) ||
                 (p.y() < titleBaseY + 11 && p.x() < leftBorder + 3) )
                return PositionTopLeft;
        }

        // Test for the top right corner
        if ( p.x() > rightBorder - 11 ) {
            if ( (p.y() < titleBaseY + 3 && p.x() > rightBorder - 11) ||
                 (p.y() < titleBaseY + 6 && p.x() > rightBorder - 6) ||
                 (p.y() < titleBaseY + 11 && p.x() > rightBorder - 3) )
                return PositionTopRight;
        }

        // Test for the top border
        if ( p.y() <= 3 || (p.y() <= titleBaseY+3 &&
                            (p.x() < captionRect.left() || p.x() > captionRect.right()) ) )
            return PositionTop;

        // The cursor must be over the center of the titlebar.
        return PositionCenter;
    }

    // Test the sides
    else if ( p.y() < bottomBorder ) {
        // Test for the left side
        if ( p.x() < leftBorder ) {
            if ( p.y() < height() - bottomCornerSize )
                return PositionLeft;
            else
                return PositionBottomLeft;
        }

        // Test for the right side
        else if ( p.x() > rightBorder ) {
            if ( p.y() < height() - bottomCornerSize )
                return PositionRight;
            else
                return PositionBottomRight;
        }

        // The cursor must be over the center of the window
        return PositionCenter;
    }

    // Test the grab bar / bottom border
    else {
        // Test for the bottom left corner
        if ( p.x() < bottomCornerSize )
            return PositionBottomLeft;

        // Test for the bottom right corner
        else if ( p.x() > width() - bottomCornerSize - 1 )
            return PositionBottomRight;

        // The cursor must be over the bottom border
        return PositionBottom;
    }

    // We should never get here
    return PositionCenter;
}



void MandrakeClient::resize( const QSize& s )
{
	widget()->resize( s );
}


void MandrakeClient::borders( int& left, int& right, int& top, int& bottom ) const
{
	int titleBarHeight     = clientHandler->titleBarHeight( );
	int grabBarHeight      = clientHandler->grabBarHeight();
	int leftBorderWidth    = clientHandler->tile( BorderLeft, isActive() )->width();
	int rightBorderWidth   = clientHandler->tile( BorderRight, isActive() )->width();

	left   = leftBorderWidth;
	right  = rightBorderWidth;
	top    = titleBarHeight;
	bottom = grabBarHeight;


	if ( ( maximizeMode() & MaximizeHorizontal ) && !options()->moveResizeMaximizedWindows())
		left = right = 0;
	if( ( maximizeMode() & MaximizeVertical ) )
	  {
	    top = clientHandler->titleBarHeight( );
	    if( !options()->moveResizeMaximizedWindows())
	      bottom = 0;
	  }
}


QSize MandrakeClient::minimumSize() const
{
	return widget()->minimumSize();
}


bool MandrakeClient::eventFilter( QObject* o, QEvent* e )
{
	if ( o != widget() )
		return false;

	switch ( e->type() )
	{
		case QEvent::Resize:
			resizeEvent( static_cast< QResizeEvent* >( e ) );
			return true;

		case QEvent::Paint:
			paintEvent( static_cast< QPaintEvent* >( e ) );
			return true;

		case QEvent::MouseButtonDblClick:
			mouseDoubleClickEvent( static_cast< QMouseEvent* >( e ) );
			return true;

		case QEvent::MouseButtonPress:
			processMousePressEvent( static_cast< QMouseEvent* >( e ) );
			return true;

		default:
	    		return false;
	}
}


}; // namespace Mandrake


extern "C"
{
	KDecorationFactory *create_factory()
	{
	  Mandrake::clientHandler = new Mandrake::MandrakeHandler();
                return Mandrake::clientHandler;
	}

}

