/*
 * Mandrake KWin client based on Keramik kwin client
 *
 * Copyright (C) 2003-2004 Laurent Montel <lmontel@mandrakesoft.com>
 * Copyright (C) 2002 Fredrik Höglund <fredrik@kde.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the license, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#ifndef __MANDRAKE_H
#define __MANDRAKE_H

#include <qbutton.h>
#include <kdecoration.h>
#include <kdecorationfactory.h>


#include "tiles.h"

class QSpacerItem;


namespace Mandrake {

    enum TilePixmap  { TitleLeft=0, TitleCenter, TitleRight,
                       GrabBarLeft, GrabBarCenter, GrabBarRight,
                       BorderLeft, BorderRight, BorderBottomLeft,
                       BorderBottomRight, LeftBorderTitle, NumTiles };

    enum Button      { MenuButton = 0, OnAllDesktopsButton, HelpButton, MinButton,
                       MaxButton, CloseButton, NumButtons,  };

    enum ButtonDeco  { Sticky = 0, Unsticky, Help, Iconify, Maximize,
                       Restore, Close,   CloseInactive, RestoreInactive, MaximizeInactive, IconifyInactive, HelpInactive,NumButtonDecos };

    struct SettingsCache
    {
        QColor  aTitleColor;
        QColor  iTitleColor;
        QColor  aTitleBlend;
        QColor  iTitleBlend;
        QColor  buttonColor;
        QString buttonsLeft;
        QString buttonsRight;
        bool showTooltips:1;
    };

    class MandrakeHandler : public KDecorationFactory {
    public:
        MandrakeHandler();
        ~MandrakeHandler();
	virtual bool reset( unsigned long changed );
	virtual KDecoration* createDecoration( KDecorationBridge* );

        bool showAppIcons() const        { return showIcons; }
        bool useShadowedText() const     { return shadowedText; }


        int titleBarHeight() const {
            return ( activeTiles[TitleCenter]->height() );
        }


        int grabBarHeight() const
            { return activeTiles[GrabBarCenter]->height(); }

        const QPixmap *roundButtonNormal() const  { return titleButtonRoundNormal; }
        const QPixmap *roundButtonHover() const  { return titleButtonRoundHover; }
        const QPixmap *roundButtonDown() const  { return titleButtonRoundDown; }

        const QPixmap *roundButtonNormalInactive() const  { return titleButtonRoundNormalInactive; }
        const QPixmap *roundButtonHoverInactive() const  { return titleButtonRoundHoverInactive; }
        const QPixmap *roundButtonDownInactive() const  { return titleButtonRoundDownInactive; }

        const QPixmap *buttonDeco( ButtonDeco deco ) const
            { return buttonDecos[ deco ]; }

        inline const QPixmap *tile( TilePixmap tilePix, bool active ) const;


    private:
        void readConfig();
        void createPixmaps();
        void destroyPixmaps();
	void addWidth  (int width,  QPixmap *&pix, bool left, QPixmap *bottomPix);
	void addHeight (int height, QPixmap *&pix);

        void flip( QPixmap *&, QPixmap *& );
        void flip( QPixmap *& );
        void pretile( QPixmap *&, int, Qt::Orientation );
        QPixmap *composite( QImage *, QImage * );
        QImage  *loadImage( const QString &, const QColor & );
        QPixmap *loadPixmap( const QString &, const QColor & );

    private:
        bool showIcons:1, shadowedText:1;
        SettingsCache *settings_cache;
        MandrakeImageDb *imageDb;

        QPixmap *activeTiles[ NumTiles ];
        QPixmap *inactiveTiles[ NumTiles ];
        QPixmap *buttonDecos[ NumButtonDecos ];

        QPixmap *buttonMenuDecos;

        QPixmap *titleButtonRoundNormal;
        QPixmap *titleButtonRoundHover;
        QPixmap *titleButtonRoundDown;

        QPixmap *titleButtonRoundNormalInactive;
        QPixmap *titleButtonRoundHoverInactive;
        QPixmap *titleButtonRoundDownInactive;


    }; // class MandrakeHandler

    class MandrakeClient;
    class MandrakeButton : public QButton
    {
    public:
        MandrakeButton( MandrakeClient *, const char *, Button, const QString & );
        ~MandrakeButton();

        int lastButton() const { return lastbutton; }

    protected:
        void enterEvent( QEvent * );
        void leaveEvent( QEvent * );
        void mousePressEvent( QMouseEvent * );
        void mouseReleaseEvent( QMouseEvent * );
    private:

        void drawButton( QPainter * );
        void drawBackgroundButton( QPainter *p, bool drawDeco=true );


    protected:
        MandrakeClient *client;
        Button button;
        bool hover;
        int lastbutton;
    }; // class MandrakeButton

    class MandrakeMenuButton : public MandrakeButton
    {
    public:
        MandrakeMenuButton( MandrakeClient *, const char *, Button, const QString & );
        ~MandrakeMenuButton();
        QPixmap *menuIcons();
        void iconChange();
    private:
        void enterEvent( QEvent * );
        void leaveEvent( QEvent * );
        void mousePressEvent( QMouseEvent * );
        void mouseReleaseEvent( QMouseEvent * );
        void drawButton( QPainter * );
        void drawBackgroundButton( QPainter *p );
        QPixmap       *activeIcon, *inactiveIcon;
    }; // class MandrakeButton



    class MandrakeClient : public KDecoration
    {
        Q_OBJECT

    public:

        MandrakeClient( KDecorationBridge* bridge, KDecorationFactory* factory );
        ~MandrakeClient();
	virtual void init();
	virtual void reset( unsigned long changed );
	virtual Position mousePosition( const QPoint& p ) const;
	virtual void borders( int& left, int& right, int& top, int& bottom ) const;
	virtual void resize( const QSize& s );
	virtual QSize minimumSize() const;
	virtual bool eventFilter( QObject* o, QEvent* e );
	virtual void activeChange();
	virtual void captionChange();
	virtual void maximizeChange();
	virtual void desktopChange();
	virtual void shadeChange() {};

    private:
	void createLayout();
        void addButtons( QBoxLayout*, const QString & );
        void updateMask(); // FRAME
        void updateCaptionBuffer();
        void iconChange();
        void stickyChange( bool );
        void resizeEvent( QResizeEvent *);// FRAME
        void paintEvent( QPaintEvent *);// FRAME
        void mouseDoubleClickEvent( QMouseEvent * );// FRAME
        int calculateLeftButtonWidth(const QString &s)const;

	int width() const { return widget()->width(); }
	int height() const { return widget()->height(); }

        void calculateCaptionRect();

        inline bool maximizedVertical() const {
            return (maximizeMode() & MaximizeVertical);
        }

private slots:
        void menuButtonPressed();
        void slotMaximize();

    private:
        QSpacerItem  *titlebar;
        MandrakeButton *button[ NumButtons ];
        MandrakeMenuButton *menuButton;
        QRect          captionRect;
        QPixmap        captionBuffer;
        QPixmap       *activeIcon, *inactiveIcon;
        bool           captionBufferDirty:1, maskDirty:1;
    }; // class MandrakeClient

} // namespace Mandrake

#endif
